"""Functions used for class registration."""
import logging
import re
from ..decorator import shared_wrappers


def dispatch(cls, n_dispatch):
    """Dispatch wrapper."""
    n_dispatch = 0 if n_dispatch is None else n_dispatch
    if n_dispatch == 0:
        logging.warning('No dispatching as `n_dispatch` is null.')
        return
    elif not isinstance(n_dispatch, int):
        raise ValueError('`n_dispatch` should be an integer')
    base_classname = make_wrapper_classname(
        cls._Wrapper__wrapped_class.__name__)
    max_num, cls_name = (
        re.search(r'Wrapped([0-9]*)\((.*)\)', base_classname).groups()
    )
    for i in range(n_dispatch):
        new_classname = (
            'Wrapped{}({})'.format(int(max_num) + i, cls_name)
        )
        cls_copy = type(new_classname,
                        cls.__bases__,
                        dict(cls.__dict__))
        register(cls_copy, verbose=False)


def assign(cls, verbose=False):
    """Assign wrapper."""
    classname = cls if isinstance(cls, str) else cls.__name__
    if classname not in shared_wrappers:
        ValueError('{} is not a registered class'.format(classname))
    if shared_wrappers[classname]._Wrapper__assigned:
        ValueError('{} is already assigned'.format(classname))
    if verbose:
        print('Assigning: {}'.format(classname))
    shared_wrappers[classname]._Wrapper__assigned = True


def unassign(cls, verbose=False):
    """Unassign wrapper."""
    classname = cls if isinstance(cls, str) else cls.__name__
    if classname not in shared_wrappers:
        ValueError('{} is not a registered class'.format(classname))
    if not shared_wrappers[classname]._Wrapper__assigned:
        logging.warning('{} is not assigned'.format(classname))
        return
    if verbose:
        print('Unassigning: {}'.format(classname))
    shared_wrappers[classname]._Wrapper__assigned = False


def unassign_all(cls, verbose=False):
    """Unassign all wrappers."""
    classname = cls if isinstance(cls, str) else cls.__name__
    if verbose:
        print('Unassigning all wrappers of {}'.format(classname))
    for classname in get_unassigned_wrappers_classnames(classname):
        unregister(classname, verbose=False)


def get_unassigned_wrappers_classnames(classname):
    """Get unassigned wrappers classnames."""
    registered_wrappers = get_registered_wrappers_classnames()
    wrappers = (
        [(wrapper_name, shared_wrappers[wrapper_name])
         for wrapper_name in registered_wrappers]
    )
    unassigned_wrappers = (
        [wrapper_name for (wrapper_name, wrapper) in wrappers
         if not wrapper._Wrapper__assigned]
    )
    unassigned_wrappers = [
        wrapper for wrapper in unassigned_wrappers
        if re.search(r'Wrapped([0-9]*)\((.*)\)', wrapper).groups() is not None
    ]
    if len(unassigned_wrappers) == 0:
        raise ValueError('No assigned wrapper found.')

    return unassigned_wrappers


def get_first_unassigned_wrapper(classname):
    """Get first unassigned wrapper."""
    unassigned_wrappers = get_unassigned_wrappers_classnames(classname)
    nums = [re.search(r'Wrapped([0-9]*)\((.*)\)', wrapper).groups()[0]
            for wrapper in unassigned_wrappers]
    nums = [int(num) for num in nums if num != '']
    min_num = sorted(nums)[0]
    wrapper_classname = 'Wrapped{}({})'.format(min_num, classname)
    return shared_wrappers[wrapper_classname]


def register(cls, verbose=False):
    """Register class."""
    classname = cls.__name__
    if verbose:
        print('Registering: {}'.format(classname))
    if classname in shared_wrappers and 'Wrapped(' not in classname:
        raise ValueError('{} is already a registered class.'.format(classname))
    shared_wrappers[classname] = cls


def unregister(cls, verbose=False):
    """Unregister class."""
    classname = cls if isinstance(cls, str) else cls.__name__
    if verbose:
        print('Unregistering: {}'.format(classname))
    if classname in shared_wrappers:
        del shared_wrappers[classname]
    else:
        logging.warning('{} is not a registered class'.format(classname))


def unregister_all():
    """Unregister all classes."""
    for classname in get_registered_wrappers_classnames():
        unregister(classname, verbose=False)


def get_registered_wrappers_classnames():
    """Get classnames of all registered wrappers."""
    return list(set([k for k in shared_wrappers if 'Wrapped' in k]))


def make_wrapper_classname(classname):
    """Return new wrapper classname based on registered classes."""
    registered_wrappers = get_registered_wrappers_classnames()
    searches = [re.search(r'Wrapped([0-9]*)\((.*)\)', elt)
                for elt in registered_wrappers if 'DWrapped' not in elt]
    nums = [search.groups()[0] for search in searches
            if search is not None and search.groups() is not None]
    if len(nums) == 0:
        wrapper_classname = 'Wrapped({})'.format(classname)
    else:
        if len(nums) == 1 and nums[0] == '':
            max_num = 2
            wrapper_classname = 'Wrapped{}({})'.format(max_num, classname)
        else:
            nums = [int(num) for num in nums if num != '']
            max_num = sorted(nums)[-1] + 1
            wrapper_classname = 'Wrapped{}({})'.format(max_num, classname)
    return wrapper_classname



# Parser utils
import os
import yaml
from ast import literal_eval
from os.path import abspath, dirname, isfile, join, realpath

dir_path = dirname(realpath(__file__))
PROJECT_DIR = abspath(join(dir_path, '..'))


def parse_config():
    """Parse configuration file and return a dictionary."""
    with open(abspath(join(PROJECT_DIR, 'config.yml')), 'r') as file:
        config = yaml.safe_load(file)
    for k, v in config.items():
        config[k] = literal_eval(v)
    return config

CONFIG = parse_config()



# PyDeco.
#
# Author: Paul Roujansky <paul@roujansky.eu>
# Last Change: June 28, 2019
# URL: https://github.com/paulroujansky/pydeco

"""The main module of the `pydeco` package."""

# Library modules
from .decorator import Decorator, MethodsDecorator
from .utils.misc import PYTHON_VERSION

# Semi-standard module versioning.
__version__ = '0.362'



"""Decorator for class methods (compatible with Python 3.0 or higher).

The herebelow class :class:`MethodsDecorator` enables to decorate multiple
methods of a given class with custom decorators by simply decorating said class
with it. One then needs to specify, as a dictionary, which decorator is used to
decorate which methods.

"""
import logging
import re
from abc import abstractmethod
from copy import deepcopy
from functools import wraps

from .utils import CONFIG, is_wrapped

global shared_wrappers
shared_wrappers = globals()

from .utils.register import (assign, dispatch,  # noqa: E402
                             get_first_unassigned_wrapper,
                             make_wrapper_classname, register)


class Decorator(object):
    """Decorator base class."""

    def __init__(self, *args, **kwargs):
        self.instances = []

    def flush_instances(self):
        """Flush instances."""
        self.instances = []

    @abstractmethod
    def wrapper(self, instance, func, *args, **kwargs):
        """Wrap func."""
        pass

    def activate(self):
        """Activate decorator for all methods of decorated instances."""
        for instance in self.instances:
            instance.activate_decorator(self.__class__.__name__)

    def deactivate(self):
        """Deactivate decorator for all methods of decorated instances."""
        for instance in self.instances:
            instance.deactivate_decorator(self.__class__.__name__)

    def is_active(self, instance):
        """Return True if decorator is active for input instance."""
        if instance not in self.instances:
            err = ('Current decorator does not decorate a method of input '
                   'class instance.')
            raise ValueError(err)
        if not is_wrapped(instance):
            return True
        else:
            return instance.is_decorator_active(self.__class__.__name__)

    def __call__(self, func):
        """Call."""
        def wrapped_func(instance, func, *args, **kwargs):
            """Call wrapper is decorator is active, otherwise call func."""
            if instance not in self.instances:
                self.instances.append(instance)

            if self.is_active(instance):
                # active decorator for the current func: wrap it
                # check if input is a bound method or not
                if hasattr(func, '__self__'):

                    # wrap func as it is a bound method
                    def func_(instance, *args, **kwargs):
                        return func(*args, **kwargs)

                    return self.wrapper(instance, func_, *args, **kwargs)
                else:
                    return self.wrapper(instance, func, *args, **kwargs)
            else:
                # inactive decorator for the current func
                return func(instance, *args, **kwargs)

        if hasattr(func, '__self__'):
            # input object is a method of an instance
            instance = func.__self__

            @wraps(func)
            def _wrapped_func(*args, **kwargs):
                return wrapped_func(instance, func, *args, **kwargs)
            return _wrapped_func

        else:
            @wraps(func)
            def _wrapped_func(instance, *args, **kwargs):
                return wrapped_func(instance, func, *args, **kwargs)
            return _wrapped_func


class MethodsDecorator(object):
    """Class that enables to decorate specific methods with given decorator.

    Parameters
    ----------
    mapping : dict
        Mapping containing decorators as key and methods (as str or a list of
        str) as values (ex: ``{Timer(): ['fit', 'predict']}``)

    Examples
    --------
    We assume here that a decorator named :class:`Timer` (used to compute
    running time of called functions) is already implemented. The below code
    applies such a decorator to the method :meth:`method_1` of class
    :class:`MyClass` so as to compute running time every time :meth:`method_1`
    is called.

    >>> @MethodsDecorator(mapping={Timer(): 'method_1'})
    >>> class MyClass():
    >>>
    >>>    def method_1(self, *args, **kwargs):
    >>>       return

    The "@" syntax is a proxy for the below implementation:

    >>> class MyClass():
    >>>
    >>>    def method_1(self, *args, **kwargs):
    >>>       return
    >>>
    >>> MyClass = MethodsDecorator(mapping={Timer(): 'method_1'})(MyClass)

    See the examples section for more insights on how to use
    :class:`MethodsDecorator`.

    """

    def __init__(self, mapping={}):

        self.mapping = mapping
        for decorator, methods in mapping.items():
            if not isinstance(methods, (tuple, list)):
                methods = [methods]
            assert callable(decorator)
            assert all([isinstance(method, str) for method in methods])
            self.mapping[decorator] = methods
        self.original_methods = dict()

    def __call__(self, cls):
        """Return wrapped input class with decorated methods."""
        mapping = self.mapping
        original_methods = self.original_methods

        class MC(type):
            """Decorating methods for the input class with given decorator.

            Attributes
            ----------
            decorated : bool
                Indicates that current class is decorated.
            class_decorator : type
                Type of the class decorator used to decorate the current class.

            """

            def __init__(cls_, name, bases, dict):
                for decorator, methods in self.mapping.items():
                    for method in methods:
                        if not hasattr(cls_, method):
                            err = 'Input class has not method "{}"'.format(
                                method)
                            raise ValueError(err)
                        if method not in self.original_methods:
                            self.original_methods[method] = (
                                getattr(cls_, method)
                            )
                        setattr(cls_, method,
                                decorator(getattr(cls_, method)))

                super(MC, cls_).__init__(name, bases, dict)

        global Wrapper

        class Wrapper(cls, metaclass=MC):
            """Wrapped class where each specified method is decorated."""

            __wrapped_class = cls  # base class
            __decorated = True  # indicates that the class is decorated
            __wrapper = self.__class__  # type of class decorator
            __decorator_mapping = mapping  # decorator mapping
            __original_methods = original_methods
            __assigned = False

            def __init__(self, *args, **kwargs):
                self._decorator_mapping = {
                    k: v for k, v in self.__decorator_mapping.items()
                }
                self.active_decorators = {
                    decorator: True for decorator in self.decorators
                }
                self.__class__.__assigned = True
                cls.__init__(self, *args, **kwargs)

            @property
            def decorators(self):
                """Return decorators."""
                return {
                    decorator.__class__.__name__: decorator
                    for decorator in self._decorator_mapping.keys()
                }

            def __deepcopy__(self, memo=None, _nil=[]):
                """Deepcopy."""
                # Remove decorators from self
                cls_self = self.__class__
                tmp_methods = dict()
                tmp_mapping = dict()
                tmp_active_decorators = dict()

                for decorator, methods in self._decorator_mapping.items():
                    tmp_mapping[decorator] = methods
                for decorator, is_active in self.active_decorators.items():
                    tmp_active_decorators[decorator] = is_active

                self._decorator_mapping = dict()
                self.active_decorators = dict()

                for method_name, method in self.__original_methods.items():
                    tmp_methods[method_name] = method
                    delattr(cls_self, method_name)

                # Copy self
                c_self = cls_self.__new__(cls_self)
                memo[id(self)] = c_self
                for k, v in self.__dict__.items():
                    setattr(c_self, k, deepcopy(v, memo))

                # Copy self's Wrapper
                cls_c_self = get_first_unassigned_wrapper(cls.__name__)
                assign(cls_c_self)
                c_self.__class__ = cls_c_self

                # Add back decorators to both self and its copy
                for method_name, method in tmp_methods.items():
                    setattr(cls_self, method_name, method)
                    setattr(cls_c_self, method_name, method)

                for decorator, methods in tmp_mapping.items():
                    decorator_name = decorator.__class__.__name__
                    c_decorator = deepcopy(decorator)
                    self._decorator_mapping[decorator] = methods
                    c_self._decorator_mapping[c_decorator] = methods

                    is_decorator_active = (
                        tmp_active_decorators[decorator_name]
                    )
                    self.active_decorators[decorator_name] = (
                        is_decorator_active)
                    c_self.active_decorators[decorator_name] = (
                        is_decorator_active)

                    for method_name in methods:
                        if not hasattr(self, method_name):
                            err = 'Input class has not method "{}"'.format(
                                method_name)
                            raise ValueError(err)
                        setattr(
                            cls_self, method_name,
                            decorator(getattr(cls_self, method_name)))
                        if not hasattr(c_self, method_name):
                            err = 'Input class has not method "{}"'.format(
                                method_name)
                            raise ValueError(err)
                        setattr(
                            cls_c_self, method_name,
                            c_decorator(getattr(cls_c_self, method_name)))
                # return copy
                return c_self

            def _check_decorator_name(self, name):
                if name not in self.decorators:
                    err = ('Could not find decorator "{}". Available '
                           'decorators: {}'.format(
                               name, list(self.decorators.keys())))
                    raise ValueError(err)

            def is_decorator_active(self, name):
                """Check if input decorator is active."""
                self._check_decorator_name(name)
                return self.active_decorators[name]

            def activate_decorator(self, name):
                """Activate decorator."""
                self._check_decorator_name(name)
                self.active_decorators[name] = True

            def deactivate_decorator(self, name):
                """Deactivate decorator."""
                self._check_decorator_name(name)
                self.active_decorators[name] = False

        # Updating wrapped class name and documentation
        Wrapper.__name__ = make_wrapper_classname(cls.__name__)
        Wrapper.__doc__ = cls.__doc__

        # Registering newly created Wrapper class
        register(Wrapper)
        # Dispatch Wrapper for multiprocessing purpose
        dispatch(Wrapper, n_dispatch=CONFIG['N_DISPATCH'])

        return Wrapper



#! /usr/bin/env python

# Standard library modules.
from setuptools import setup, find_packages
import codecs
import os
import re
import sys

# platform specific dependency
exclude = ['data', 'contrib', 'doc', 'tests']

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


def get_absolute_path(*args):
    """Transform relative pathnames into absolute pathnames."""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), *args)


def get_contents(*args):
    """Get the contents of a file relative to the source distribution dir."""
    with codecs.open(get_absolute_path(*args), 'r', 'UTF-8') as handle:
        return handle.read()


def get_version(*args):
    """Extract the version number from a Python module."""
    contents = get_contents(*args)
    metadata = dict(re.findall('__([a-z]+)__ = [\'"]([^\'"]+)', contents))
    return metadata['version']


setup(
    name='pydeco',
    version=get_version('pydeco', '__init__.py'),
    python_requires='>=3',
    description='Python Class Methods Decorator',
    url='https://github.com/paulroujansky/pydeco',
    author='Paul Roujansky',
    long_description=get_contents('README.rst'),
    long_description_content_type='text/x-rst',
    author_email='paul@roujansky.eu',
    license='MIT',
    test_suite='pydeco.tests',
    install_requires=requirements,
    packages=find_packages(exclude=exclude),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False)
